#include "inkey.ch"
#include "fileio.ch"

STATIC ObjectsArr := {}, lChanged := .F., ClrSpec

////////////////////////////////////////////////////////////////////////////
//
//  Function : ColorPick()
//
//  Purpose  : To display colors of items with their foreground and
//             background descriptions and allow the user to change them.
//
//  Author   : Jake Jacob
//
//  Returns  : The ID of the SubArray the user was on upon termination.
//
//  Syntax   : ColorPick( [<nPaletteID>], [<BoxColorStrID>], [<lSortArray>] )
//
//
//
//             <nPaletteID>    = Palette ID of the color system. <nPaletteID>
//                               points to the color array stored in that
//                               Palette.
//
//
//             <BoxColorStrID> = Character or Numeric. If Character then, box
//                               border color and column headings will be set
//                               accordingly. If Numeric then, ColorPick()
//                               will search all the SubArrays to find
//                               a matching ID and set the border color and
//                               the column headings according to that ID's
//                               color string.
//
//                               If not specified or no matching ID is found
//                               then default color setting will be "GR+/B*".
//
//
//
//             <lSortArray>    = Logical. If TRUE then ColorPick() will sort
//                               <aItemsArray> according to the first element
//                               of each SubArray which is the Item's name.
//
//
//
//  Usage    : Because ColorPick() terminates after each color change,
//             it is nessecary to put it inside a loop that executes it
//             continueously until a desired termination condition is
//             reached, then it is recommended to execute ColorPick()
//             one last time without any parameters to set the internal
//             STATIC variables to NIL.
//
//             Example:
//
//                   LOCAL nItemID
//                   
//                   
//
//                   DO WHILE LASTKEY() <> K_ESC .AND. LASTKEY() <> K_RETURN
//
//                     nItemID := ColorPick( nPaletteID,, .T. )
//
//                     SETCOLOR( ColorGet( nPaletteID, nItemID ) )
//
//                     < code to refresh the screen >
//
//                   ENDDO
//
//                   ColorPick()
//
//                   
//                   
//                   
//
//             ColorPick() terminates thusly to allow for live color display
//             changes to the individual items -outside the function- on the
//             screen while the user is changing them. If this is not the
//             desired effect then, use ColorSet().
//
//
//  Note     : ColorPick()'s border box will resize itself to include at
//             least the Item and the Foreground columns. The Background
//             column can be panned into if it does not appear readily.
//
//             "W+/GR*" (Bright White on Yellow) and "W+/W+" (Bright White
//             on Bright White) may not appear properly on the screen while
//             the function is executing. This is attributed to a minor
//             limitation in ColorPick(). However, the internal color strings
//             of the array will be set properly and should have no effect
//             on the color display outside the function. This is only a
//             minor aesthetic annoyance.
//
//             Note, that ColorPick() sets SETBLINK( .F. ), and will leave
//             it set as such even after it terminates.
//
//             
//
//  Copyright (C) 1990 Nantucket Corp. 
//
////////////////////////////////////////////////////////////////////////////

FUNCTION ColorPick( nPaletteID, ColorStr, lSortArray )

STATIC O, X, Y, Z, n := 1, Width := 0, T, L, B, R, nPID := 0

LOCAL i := 0, j := 0, sOldColor, OldRow, El_Type, Element, aPalette
LOCAL iKey := 0, k, TableBlock, Flag := .F., OldCol, ColorArray
LOCAL SubScript, q := 1, BackColor, Str, OldCurSet, O_Flag := .F.


OldCurSet := SETCURSOR( 0 )


IF SETBLINK()

  SETBLINK( .F. )

ENDIF


// if no parameters are passed then assign all the static vars to NIL

IF PCOUNT() == 0

   O := X := Y := Z := NIL
   n := 1
   SETCURSOR( OldCurSet )
   lChanged := .T.
   RETURN NIL

ENDIF


nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
   @ 0,0
   @ 0,0 SAY 'Palette ID does not exist...'
   @ 1,0
   @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
             ALLTRIM( STR( PROCLINE(1) ) ) + ')'
   QUIT
ENDIF

aPalette   := Palette( nPaletteID )
ColorArray := aPalette[3]


IF O == NIL

   ClrSpec := ''

   FOR i := 1 TO 16
      FOR j := 1 TO 16
         ClrSpec += ForeArray()[ i ] + '/' + BackArray()[ j ] + ','
      NEXT
   NEXT

   ClrSpec := SUBSTR( ClrSpec, 1, LEN( ClrSpec ) - 1 )

ENDIF


IF nPID <> nPaletteID .OR. lChanged

  
   nPID := nPaletteID
   lChanged := .F.
   O := X := Y := Z := NIL
   n := 1


   // save the cursor position.

   OldRow := ROW()
   OldCol := COL()


   // if ColorArray contains no elements, return NIL

   IF LEN( ColorArray ) == 0

      RETURN ''

   ELSE


      // check if ColorArray contains SubArrays...and that each SubArray contains
      // 3 elements.

      Element := ASCAN(ColorArray, {|SubArray| ;
                             Flag := .NOT. ( VALTYPE( SubArray ) == 'A' ;
                             .AND. LEN( SubArray ) == 3 ) } )


      IF Flag

         @ 0,0
         @ 0,0 SAY 'Color SubArray ' + ALLTRIM( STR( Element ) ) + ' contains' + ;
                   ' illegal elements...'
         @ 1,0
         @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
                   ALLTRIM( STR( PROCLINE(1) ) ) + ')'
         QUIT

      ELSE


         // check the type of each element of each SubArray.

         FOR i := 1 TO LEN(ColorArray)


            // if Item name is not char then set Flag to true.

            El_Type := VALTYPE( ColorArray[ i, 1 ] )
            Flag := .NOT. ( El_Type == 'C' )


            // if Color string is not char or NIL then set Flag to true.

            IF .NOT. Flag
               El_Type := VALTYPE( ColorArray[ i, 2 ] )
               Flag := ( .NOT. ( El_Type == 'C' ) ) .AND. .NOT. ( El_Type == 'U' )
            ENDIF


            // if ID is not numeric then set Flag to true.

           IF .NOT. Flag
              El_Type := VALTYPE( ColorArray[ i, 3 ] )
              Flag := .NOT. ( El_Type == 'N' )
           ENDIF


           IF Flag

              @ 0,0
              @ 0,0 SAY 'Color SubArray ' + ALLTRIM( STR( i ) ) + ;
                        ' contains an illegal element type...'
              @ 1,0
              @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
                        ALLTRIM( STR( PROCLINE(1) ) ) + ')'
              QUIT

           ENDIF


        NEXT
  

     ENDIF
  
   ENDIF


   // if 2nd element of each sub-array (color string) is empty then assign 
   // it 'W/N'.
 
   AEVAL(ColorArray,{|SubArray| ;
         SubArray[2]:=IF(EMPTY(SubArray[2]),'W/N',ALLTRIM(SubArray[2]))})

  
   // format and check the color strings in ColorArray.

   FOR i := 1 TO LEN( ColorArray )

      // format the color strings to " <ForeGround> [+] , <BackGround> [*] "
      ColorArray[ i, 2 ] := CleanUpClr( ColorArray[ i, 2 ] )
  
      // Now, check if the color strings are legal.
      Flag :=  .NOT. CheckClrStr( ColorArray[ i, 2 ] )
  
      IF Flag
  
         @ 0,0
         @ 0,0 SAY 'Color string in Color SubArray ' + ALLTRIM( STR ( i ) ) + ;
                   ' is not legal...'
         @ 1,0
         @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
                   ALLTRIM( STR( PROCLINE(1) ) ) + ')'
         QUIT
         
      ENDIF
 
   NEXT
    

   // determine the width of the largest Item name in ColorArray.

   AEVAL( ColorArray, {|Str| Width := MAX( Width, LEN( Str[ 1 ] ) ) } )
   Width := IF( Width < 4, 4, IF( Width > 55, 55, Width ) )
 
 
   // if lSortArray is true then sort ColorArray according to Item name.
 
   IF lSortArray <> NIL .AND. lSortArray
      ASORT(ColorArray,,, {| x, y | x[1] < y[1] } )
   ENDIF
   
 
   T := aPalette[ 2, 1 ]
   L := aPalette[ 2, 2 ]
   B := aPalette[ 2, 3 ]
   R := aPalette[ 2, 4 ]
 
 
   // resize the TBrowse box coordinates.
 
   IF T == NIL
      IF LEN( ColorArray ) + 4 > MAXROW()
         T := 0
         B := MAXROW()
      ELSE
         T := INT( MAXROW() / 2 ) - INT( LEN( ColorArray ) / 2 )
         B := L + LEN( ColorArray ) + 3
      ENDIF
   ENDIF
 
   IF L == NIL
      IF Width + 30 > MAXCOL()
         L := INT( MAXCOL() / 2 ) - INT( ( Width + 15 ) / 2 )
         R := L + Width + 18
      ELSE
         L := INT( MAXCOL() / 2 ) - INT( ( Width + 30 ) / 2 )
         R := L + Width + 33
      ENDIF
   ENDIF
 
   B := MIN( T + LEN( ColorArray ) + 3, B )
 
   IF R - L - 3 < Width + 30
      R := L + Width + 18
   ELSE
      R := L + Width + 33
   ENDIF
 
 
   // set the TBrowse box and column header colors if ColorStr is NIL.

   IF ColorStr == NIL
      ColorStr  := 'GR+/B*'
 
 
   // if an ID was passed instead of a color string then search for a matching
   // color string in ColorArray.
  
   ELSEIF VALTYPE( ColorStr ) == 'N'
 
 
      // if ID is not found then assign it to the default color string.
 
      IF EMPTY( ColorStr := ColorGet( nPaletteID, ColorStr ) )
         ColorStr := 'GR+/B*'
      ENDIF
 
 
   ENDIF
 
 
   ColorStr := CleanUpClr( ColorStr )
 
 
   // find the background color of the box and header.
 
   BackColor := STRTRAN( BackColor( ColorStr ), '*', '+' ) + '/N'
   sOldColor := SETCOLOR( 'N/N' )
   @ T, L CLEAR TO B, R
   sOldColor := SETCOLOR( BackColor )
 
 
   // draw the TBrowse box
 
   @ T, L, B, R BOX ''
   @ T+1, L+1 SAY ''
   @ T+2, L+1 SAY ''
   @ T+1, R-1 SAY ''
   @ T+2, R-1 SAY ''
   
 
   // create the TBrowse object.
 
   O := TBrowseNew( T + 1, L + 2, B - 1, R - 2 )
 
   O:HeadSep := ''
   
 
   // TBrowse has to have a SkipBlock.
 
   O:SkipBlock := {|x| n += ( k := Skipper( n, x, LEN(ColorArray) ) ), k }
 
 
   // and these suckers too...geeez
 
   O:GoTopBlock := {|| n := 1 }
   O:GoBottomBlock := {|| n := LEN(ColorArray) }
 
 
   // define a color table-string by determinning all the possible
   // color combinations and putting them all into O:ColorSpec.
 
   O:ColorSpec := ClrSpec
   O:ColorSpec := BackColor + ',' + O:ColorSpec
 
 
   // create the 1st column.
 
   X := TBColumnNew( 'ITEM', ;
        {|| LEFT( ColorArray[ n, 1 ] + SPACE( Width ), Width ) } )
   
 
   // initialize a color block to retrieve the ColorArray reference to the
   // internal color table.
 
   X:ColorBlock := {|| TablePointer( ColorArray[ n, 2 ] ) }
 
 
   // create the 2nd column which is the foreground color.
 
   Y := TBColumnNew( 'FOREGROUND', ;
      {|| LEFT( ColorText()[ ASCAN(ForeArray(),;
      ForeColor( ColorArray[n,2] ) ) ] + SPACE( 14 ), 14 ) } )
   
 
   // create the 3rd column which is the background color.
 
   Z := TBColumnNew( 'BACKGROUND', ;
      {|| LEFT( ColorText()[ ASCAN(BackArray(),;
      BackColor( ColorArray[n,2] ) ) ] + SPACE( 14 ), 14 ) } )
 
 
   // define white on black and black on white as the selected and
   // the unselected colors of the foreground and background displays.
 
   Z:ColorBlock := Y:ColorBlock := {|| { 114, 9 } }
 
 
   X:DefColor := Y:DefColor := Z:DefColor := TablePointer( UPPER(ColorStr) )
   
 
   // add those newly created columns to the TBrowse object.
   O:AddColumn( X )
   O:AddColumn( Y )
   O:AddColumn( Z )
   
 
   O:Freeze := 1
 
 
ENDIF
   
 
Flag := .T.
SETCOLOR( ColorStr )


// start the TBrowse loop

WHILE iKey != K_ESC .AND. iKey != K_RETURN


   // if last key pressed was a + or a - then toggle a flag variable that will
   // cause the termination the TBrowse loop.
 
   IF iKey == 43 .OR. iKey == 45
      O_Flag := .T.
   ENDIF
   
 
   iKey := 0
 
 
   // keep cursor away from column one.
 
   IF O:ColPos == 1
      O:Right()
   ENDIF
 
 
   // start the stabilization loop
 
   WHILE ( .NOT. O:Stable ) .AND. iKey == 0
      O:Stabilize()
 
 
      // some cosmetic re-adjustment
 
      IF Flag
         @ T+1, L+Width+2  SAY ''
         @ T+1, L+Width+17 SAY ''
         Flag := .F.
      ENDIF
    
    
      iKey := InKey()
   ENDDO
 
 
   // terminate the TBrowse loop if the last key pressed was a + or a -.
 
   IF O_Flag
      EXIT
   ENDIF
   
 
   // if nothing was pressed then we will simply wait.
 
   IF iKey == 0
      iKey := InKey(0)
   ENDIF
 
 
   // process the directional keys or the + or - keys, if pressed.
 
   IF O:Stable
   
      DO CASE
         CASE ( iKey == K_DOWN )
            O:Down()
     
         CASE ( iKey == K_UP )
            O:Up()
     
         CASE ( iKey == K_RIGHT )
            O:Right()
            Flag := .T.
     
         CASE ( iKey == K_LEFT )
            O:Left()
            Flag := .T.
     
         CASE ( iKey == K_PGDN )
            O:PageDown()
     
         CASE ( iKey == K_PGUP )
            O:PageUp()
     
         CASE ( iKey == K_HOME )
            O:GoTop()
     
         CASE ( iKey == K_END )
            O:GoBottom()
     
     
         // if the + key was pressed then choose the next color defined in
         // ColorText().
     
         CASE ( iKey == 43 )         /* '+' */
     
     
            // check if the TBrowse cursor was sitting on the foreground column.
     
            IF O:ColPos == 2
     
     
               // get the current color, locate it in ForeArray()
               // and choose the next color in the list.
      
               SubScript := ( ASCAN( ForeArray(), ;
                 ForeColor( ColorArray[ n, 2 ] ) ) % LEN( ForeArray() ) ) + 1
       
       
               // assign the chosen color to ColorArray.
       
               ColorArray[ n, 2 ] := ForeArray()[ SubScript ] + '/' + ;
                 BackColor( ColorArray[ n, 2 ] )
       
       
            // but what if the cursor was sitting on the background column.
       
            ELSE
      
      
               // get the current color, locate it in ColorText()
               // and choose the next color in the list.
       
               SubScript := ( ASCAN( BackArray(), ;
                 BackColor( ColorArray[ n, 2 ] ) ) % LEN( BackArray() ) ) + 1
       
       
               // assign the chosen color to ColorArray.
       
               ColorArray[ n, 2 ] := ForeColor( ColorArray[ n, 2 ] ) + '/' + ;
                 BackArray()[ SubScript ]
                 
       
      
            ENDIF
    
     
            // render UnStabilized.
    
            O:RefreshCurrent()
     
    
         // if the - key was pressed then choose the previous color defined in
         // ColorText().
     
         CASE ( iKey == 45 )         /* '-' */
     
    
            // check if the TBrowse cursor was sitting on the foreground column.
     
            IF O:ColPos == 2
      
     
               // get the current color, locate it in ColorText()
               // and choose the previous color in the list.
          
               SubScript := ASCAN( ForeArray(), ;
                            ForeColor( ColorArray[ n, 2 ] ) ) - 1
          
               SubScript := IF( SubScript == 0, LEN( ForeArray() ), SubScript )
          
          
               // assign the chosen color to ColorArray.
          
               ColorArray[ n, 2 ] := ForeArray()[ SubScript ] + '/' + ;
                 BackColor( ColorArray[ n, 2 ] )
          
          
            // but what if the cursor was sitting on the background column.
          
            ELSE
          
          
               // get the current color, locate it in ColorText()
               // and choose the previous color in the list.
             
               SubScript := ASCAN( BackArray(), ;
                            BackColor( ColorArray[ n, 2 ] ) ) - 1
             
               SubScript := IF( SubScript == 0, LEN( BackArray() ), SubScript )
             
             
               // assign the chosen color to ColorArray.
             
               ColorArray[ n, 2 ] := ForeColor( ColorArray[ n, 2 ] ) + '/' + ;
                 BackArray()[ SubScript ]
             
             
            ENDIF
      
            // render UnStabilized.
     
            O:RefreshCurrent()
    
    
      ENDCASE
  
  
   ENDIF
 
 
ENDDO
 
SETCOLOR( sOldColor )
SETCURSOR( OldCurSet )
@ OldRow, OldCol SAY ''

RETURN ColorArray[ n, 3 ]
**************************** END ColorPick() *******************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : ColorSet()
//
//  Purpose  : To invoke ColorPick() Continueously.
//
//  Returns  : Color Array
//
//  Syntax   : ColorSet( [<nPaletteID>], [<BoxColorStrID>], [<lSortArray>] )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION ColorSet( nPaletteID, ColorStr, lSort )

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

DO WHILE LASTKEY() <> K_ESC
  ColorPick( nPaletteID, ColorStr, lSort )
ENDDO

ColorPick()

RETURN PaletteGet( nPaletteID )
**************************** END ColorSet() *****************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : PaletteNew()
//
//  Purpose  : To create a new palette for the color system.
//
//  Returns  : The newly created Palette ID. (numeric)
//
//  Syntax   : PaletteNew( [<nTop>], [<nLeft>], [<nBottom>], [<nRight>] )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION PaletteNew( T, L, B, R )

LOCAL nPaletteID

IF PCOUNT() <> 0

  IF ( VALTYPE( T ) <> 'N' .AND. T <> NIL ) .OR.;
     ( VALTYPE( L ) <> 'N' .AND. L <> NIL ) .OR.;
     ( VALTYPE( B ) <> 'N' .AND. B <> NIL ) .OR.;
     ( VALTYPE( R ) <> 'N' .AND. R <> NIL )

    @ 0,0
    @ 0,0 SAY 'Illegal box parameters...'
    @ 1,0
    @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
              ALLTRIM( STR( PROCLINE(1) ) ) + ')'
    QUIT
  ENDIF

ENDIF

nPaletteID := 0

IF LEN( ObjectsArr ) <> 0

  AEVAL( ObjectsArr, {|aPalette| ;
                nPaletteID := MAX( aPalette[1], nPaletteID ) } )
ENDIF

AADD( ObjectsArr, { ++nPaletteID, { T, L, B, R }, {} } )

RETURN nPaletteID
**************************** END PaletteNew() *****************************





////////////////////////////////////////////////////////////////////////////
//
//  Function : PalettePut()
//
//  Purpose  : To replace a color array into an existing palette.
//
//  Returns  : Color array
//
//  Syntax   : PalettePut( [<nPaletteID>], [<ColorArray>] )
//
////////////////////////////////////////////////////////////////////////////

FUNCTION PalettePut( nPaletteID, ColorArray )

LOCAL Flag := .F., i, El_Type, Element

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

IF ColorArray == NIL .OR. LEN( ColorArray ) == 0

  ColorArray := {}

ELSE

  ColorArray := ACLONE( ColorArray )



  // check if ColorArray contains SubArrays...and that each SubArray contains
  // 3 elements.

  Element := ASCAN(ColorArray, {|SubArray| ;
                           Flag := .NOT. ( VALTYPE( SubArray ) == 'A' ;
                           .AND. LEN( SubArray ) == 3 ) } )


  IF Flag

    @ 0,0
    @ 0,0 SAY 'Color SubArray ' + ALLTRIM( STR( Element ) ) + ' contains' + ;
              ' illegal elements...'
    @ 1,0
    @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
              ALLTRIM( STR( PROCLINE(1) ) ) + ')'
    QUIT

  ELSE


    // check the type of each element of each SubArray.

    FOR i := 1 TO LEN(ColorArray)


      // if Item name is not char then set Flag to true.

      El_Type := VALTYPE( ColorArray[ i, 1 ] )
      Flag := .NOT. ( El_Type == 'C' )


      // if Color string is not char or NIL then set Flag to true.

      IF .NOT. Flag
        El_Type := VALTYPE( ColorArray[ i, 2 ] )
        Flag := ( .NOT. ( El_Type == 'C' ) ) .AND. .NOT. ( El_Type == 'U' )
      ENDIF


      // if ID is not numeric then set Flag to true.

      IF .NOT. Flag
        El_Type := VALTYPE( ColorArray[ i, 3 ] )
        Flag := .NOT. ( El_Type == 'N' )
      ENDIF


      IF Flag

        @ 0,0
        @ 0,0 SAY 'Color SubArray ' + ALLTRIM( STR( i ) ) + ;
                  ' contains an illegal element type...'
        @ 1,0
        @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
                  ALLTRIM( STR( PROCLINE(1) ) ) + ')'
        QUIT

      ENDIF


    NEXT
  

  ENDIF


  // if 2nd element of each sub-array (color string) is empty then assign 
  // it 'W/N'.

  AEVAL(ColorArray,{|SubArray| ;
    SubArray[2]:=IF(EMPTY(SubArray[2]),'W/N',ALLTRIM(SubArray[2]))})

  
  // format and check the color strings in ColorArray.

  FOR i := 1 TO LEN( ColorArray )

    // format the color strings to " <ForeGround> [+] , <BackGround> [*] "
    ColorArray[ i, 2 ] := CleanUpClr( ColorArray[ i, 2 ] )

    // Now, check if the color strings are legal.
    Flag :=  .NOT. CheckClrStr( ColorArray[ i, 2 ] )

    IF Flag

      @ 0,0
      @ 0,0 SAY 'Color string in Color SubArray ' + ALLTRIM( STR ( i ) ) + ;
                ' is not legal...'
      @ 1,0
      @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
                ALLTRIM( STR( PROCLINE(1) ) ) + ')'
      QUIT
      
    ENDIF

  NEXT

ENDIF

lChanged := .T.

RETURN PaletteGet( nPaletteID )
**************************** END PalettePut() *****************************







////////////////////////////////////////////////////////////////////////////
//
//  Function : PaletteDel()
//
//  Purpose  : To Delete a Palette.
//
//  Returns  : NIL
//
//  Syntax   : PaletteDel( [<nPaletteID>] )
//
////////////////////////////////////////////////////////////////////////////

FUNCTION PaletteDel( nPaletteID )

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ADEL( ObjectsArr, ASCAN( ObjectsArr,;
                         {|aPalette| aPalette[1] == nPaletteID } ) )

ASIZE( ObjectsArr, LEN( ObjectsArr ) - 1 )

RETURN NIL
**************************** END PaletteDel() *****************************






////////////////////////////////////////////////////////////////////////////
//
//  Function : PaletteBox()
//
//  Purpose  : To Assign the Palette new box coordinates, or optionally
//             return the old box settings in an array.
//
//  Syntax   : PaletteBox( [<nPaletteID>], [<nTop>], [<nleft>],
//                         [<nBottom>], [<nRight>] )
//
//  Returns  : An array which contains the old box settings.
//
////////////////////////////////////////////////////////////////////////////

FUNCTION PaletteBox( nPaletteID, T, L, B, R )

LOCAL aOldSet, aPalette

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

aPalette := Palette( nPaletteID )
aOldSet  := aPalette[ 2 ]

IF ( VALTYPE( T ) <> 'N' .AND. T <> NIL ) .OR.;
   ( VALTYPE( L ) <> 'N' .AND. L <> NIL ) .OR.;
   ( VALTYPE( B ) <> 'N' .AND. B <> NIL ) .OR.;
   ( VALTYPE( R ) <> 'N' .AND. R <> NIL )

  @ 0,0
  @ 0,0 SAY 'Illegal box parameters...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

IF PCOUNT() <> 0
  aPalette[ 2, 1 ] := IF( T == NIL, aPalette[ 2, 1 ], T )
  aPalette[ 2, 2 ] := IF( L == NIL, aPalette[ 2, 2 ], L )
  aPalette[ 2, 3 ] := IF( B == NIL, aPalette[ 2, 3 ], B )
  aPalette[ 2, 4 ] := IF( R == NIL, aPalette[ 2, 4 ], R )
  lChanged := .T.
ENDIF

RETURN aOldSet
**************************** END PaletteBox() *****************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : PaletteGet()
//
//  Returns  : Color Array
//
//  Syntax   : PaletteGet( [<nPaletteID>] )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION PaletteGet( nPaletteID )

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

RETURN ACLONE( Palette( nPaletteID )[3] )
**************************** END PaletteGet() ****************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : PaletteSave()
//
//  Purpose  : To save a Color Array into a file.
//
//  Returns  : 0 if successful.
//             1 if file creation error.
//             2 if file write error.
//
//  Syntax   : PaletteSave( [<nPaletteID>], [<cFileName>] )
//
//  Note     : If <cFileName> is not specified, then default name is
//             "PALETTE.CLR".  Note also that default extension is
//             ".CLR" unless specified. If no extension is desired then
//             specify "." period at the end of <cFileName>.
//
///////////////////////////////////////////////////////////////////////////

FUNCTION PaletteSave( nPaletteID, cFileName )

LOCAL Handle, i, ColorArray, RetValue := 0

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

cFileName  := IF( cFileName  == NIL, 'PALETTE.CLR', cFileName  )
cFileName  := IF( AT( '.', cFileName ) > 0, cFileName, cFileName + '.CLR' )

IF ( Handle := FCREATE( cFileName ) ) == -1

  RetValue := 1

ELSE

  FOR i := 1 TO LEN( ColorArray := Palette( nPaletteID )[3] )

    FWRITE( Handle, ColorArray[ i, 1 ] + ';' + ColorArray[ i, 2 ] + ';' + ;
                  ALLTRIM( STR( ColorArray[ i, 3 ] ) ) + ;
                  CHR( 13 ) + CHR( 10 ) )

    IF FERROR() <> 0
      RetValue := 2
      FCLOSE( Handle )
      EXIT
    ENDIF

  NEXT

ENDIF

IF RetValue == 0
  FCLOSE( Handle )
ENDIF

RETURN RetValue
**************************** END PaletteSave() ****************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : PaletteLoad()
//
//  Purpose  : To load a Color Array from a file.
//
//  Syntax   : PaletteLoad( [<nPaletteID>], [<cFileName>] )
//
//  Returns  : 0 if successful.
//             1 if error openning file.
//             2 if error reading file.
//             3 if file is corrupted.
//
//  Note     : If <cFileName> is not specified then "PALETTE.CLR" is
//             assumed. If no extension was specified then ".CLR" is
//             Assumed. If no extension exists then specify <cFileName>
//             with a "." period at the end of <cFileName>.
//
///////////////////////////////////////////////////////////////////////////

FUNCTION PaletteLoad( nPaletteID, cFileName )

LOCAL RetValue := 0, ColorArray := {}
LOCAL Handle, Item, ColorStr, ItemID, aPalette

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

cFileName  := IF( cFileName  == NIL, 'PALETTE.CLR', cFileName  )
cFileName  := IF( AT( '.', cFileName ) > 0, cFileName, cFileName + '.CLR' )

IF ( Handle := FOPEN( cFileName ) ) == -1

  RetValue := 1

ELSE

  FEval( Handle, {|cLine| RetValue := IF(RetValue>0,RetValue,;
                                      IF(FERROR()>0,2,;
                                      IF(LEN(cLine)-;
                                      LEN(STRTRAN(cline,';',''))<>2,3,0))),;
         Item     := IF(RetValue==0,SUBSTR(cLine,1,AT(';',cLine)-1),''),;
         ColorStr := IF(RetValue==0,SUBSTR(cLine,AT(';',cLine)+1,;
                                  RAT(';',cLine)-AT(';',cLine)-1),''),;
         ItemID   := IF(RetValue==0,VAL(SUBSTR(cLine,RAT(';',cLine)+1)),''),;
         AADD( ColorArray, { Item, ColorStr, ItemID })})

  FCLOSE( Handle )

ENDIF

aPalette      := Palette( nPaletteID )
aPalette[ 3 ] := ASIZE( ColorArray, LEN( ColorArray ) - 1 )

RETURN RetValue
**************************** END PaletteLoad() ****************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : ColorAdd()
//
//  Purpose  : To add a color to ColorArray
//
//  Returns  : Color array
//
//  Syntax   : ColorAdd( [<nPaletteID>], <nItemID>, [<cItemName>],;
//                       [<cColorString>] )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION ColorAdd( nPaletteID, nItemID, Item, ColorStr )

LOCAL sFore, sBack, ColorArray

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorArray := Palette( nPaletteID )[3]

IF PCOUNT() == 0
  @ 0,0
  @ 0,0 SAY 'No Parameters Passed...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

IF nItemID == NIL
  @ 0,0
  @ 0,0 SAY 'Color ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ELSEIF VALTYPE( nItemID ) <> 'N'
  @ 0,0
  @ 0,0 SAY 'Parameter 2 is not numeric type...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

IF Item == NIL
  Item := ''
ELSEIF VALTYPE( Item ) <> 'C'
  @ 0,0
  @ 0,0 SAY 'Parameter 3 is not character type...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

IF ColorStr == NIL .OR. EMPTY( ColorStr )
  ColorStr := 'W/N'
ELSEIF VALTYPE( ColorStr ) <> 'C'
  @ 0,0
  @ 0,0 SAY 'Parameter 4 is not character type...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

IF ASCAN( ColorArray, {|SubArray| SubArray[3] == nItemID } ) > 0
  @ 0,0
  @ 0,0 SAY 'Color ID duplication...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorStr := CleanUpClr( ColorStr )
sFore := ForeColor( ColorStr )

IF ASCAN( ForeArray(), sFore ) == 0
  @ 0,0
  @ 0,0 SAY 'Illegal ForeGround color string...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

sBack := BackColor( ColorStr )

IF ASCAN( BackArray(), sBack ) == 0
  @ 0,0
  @ 0,0 SAY 'Illegal BackGround color string...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorStr := sFore + '/' + sBack

AADD( ColorArray, { Item, ColorStr, nItemID } )
lChanged := .T.

RETURN PaletteGet( nPaletteID )
**************************** END ColorAdd() *****************************





////////////////////////////////////////////////////////////////////////////
//
//  Function : ColorGet()
//
//  Purpose  : To return a color string for a specified nItemID if 
//             nItemID resides in ColorArray, or a Nul if not found.
//
//  Returns  : Color String of the nItemID specified.
//
//  Syntax   : ColorGet( [<nPaletteID>], <nItemID> )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION ColorGet( nPaletteID, nItemID )

LOCAL SubScript, ColorArray

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorArray := Palette( nPaletteID )[3]

IF ( SubScript := IDElement( nPaletteID, nItemID ) ) == 0
  RETURN ''
ENDIF

RETURN ColorArray[ SubScript, 2 ]
**************************** END ColorGet() **********************************






///////////////////////////////////////////////////////////////////////////
//
//  Function : ColorPut()
//
//  Purpose  : Changes a singular Item name and/or its color string
//             in a Palette specified with nPaletteID.
//
//  Returns  : Color Array
//
//  Syntax   : ColorPut( [<nPaletteID>], <nItemID>, [<cItem>],
//                       [<cColorStr>] )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION ColorPut( nPaletteID, nItemID, cItem, ColorStr )

LOCAL ColorArray, SubScript

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorArray := Palette( nPaletteID )[3]

IF ( SubScript  := IDElement( nPaletteID, nItemID ) ) == 0
  @ 0,0
  @ 0,0 SAY 'Item ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorStr := CleanUpClr( ColorStr )

IF .NOT. CheckClrStr( ColorStr )
  @ 0,0
  @ 0,0 SAY 'Illegal color string...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

lChanged := .T.
ColorArray[ SubScript, 1 ] := IF( cItem == NIL, ;
                                  ColorArray[ SubScript, 1 ], cItem )
ColorArray[ SubScript, 2 ] := IF( ColorStr == NIL, ;
                                  ColorArray[ SubScript, 1 ], ColorStr )

RETURN PaletteGet( nPaletteID )
**************************** END ColorPut() *******************************





///////////////////////////////////////////////////////////////////////////
//
//  Function : ColorDelete()
//
//  Purpose  : Deletes a single Item and its color and ID from Color array.
//
//  Returns  : Color Array
//
//  Syntax   : ColorDelete( [<nPaletteID>], <nItemID> )
//
///////////////////////////////////////////////////////////////////////////

FUNCTION ColorDelete( nPaletteID, nItemID )

LOCAL ColorArray, SubScript

nPaletteID := IF( nPaletteID == NIL, ObjectsArr[ 1, 1 ], nPaletteID )

IF nPaletteID == NIL
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

ColorArray := Palette( nPaletteID )[3]
IF ( SubScript  := IDElement( nPaletteID, nItemID ) ) == 0
  @ 0,0
  @ 0,0 SAY 'Item ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  QUIT
ENDIF

lChanged := .T.
ADEL( ColorArray, SubScript )
ASIZE( ColorArray, LEN( ColorArray ) - 1 )

RETURN PaletteGet( nPaletteID )
**************************** END ColorDelete() *****************************





/////////////////////////////////////////////////////////////
//             S T A T I C    F U N C T I O N S            //
/////////////////////////////////////////////////////////////

STATIC FUNCTION Skipper( n, x, ALen )
  IF x > 0
    x := IF( n + x > ALen, ALen - n, x )
  ELSEIF x < 0
    x := IF( n + x < 1   , 1 - n   , x )
  ENDIF
RETURN x



STATIC FUNCTION Palette( nPaletteID )
LOCAL SubScript
IF VALTYPE( nPaletteID ) <> 'N'
  @ 0,0
  @ 0,0 SAY 'Palette ID is not numeric...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  @ 2,0
  @ 2,0 SAY 'Called from ' + PROCNAME(2) + '(' + ;
            ALLTRIM( STR( PROCLINE(2) ) ) + ')'
  QUIT
ENDIF
SubScript := ASCAN( ObjectsArr,;
                    {|aPalette| aPalette[1] == nPaletteID } )
IF SubScript = 0
  @ 0,0
  @ 0,0 SAY 'Palette ID does not exist...'
  @ 1,0
  @ 1,0 SAY 'Called from ' + PROCNAME(1) + '(' + ;
            ALLTRIM( STR( PROCLINE(1) ) ) + ')'
  @ 2,0
  @ 2,0 SAY 'Called from ' + PROCNAME(2) + '(' + ;
            ALLTRIM( STR( PROCLINE(2) ) ) + ')'
  QUIT
ENDIF
RETURN ObjectsArr[ SubScript ]



STATIC FUNCTION TablePointer( ColorStr )
LOCAL Pointer, sFore, sBack
sFore := ASCAN( ForeArray(), SUBSTR( ColorStr, 1, AT( '/', ColorStr ) - 1 ) )
sBack := ASCAN( BackArray(), SUBSTR( ColorStr, AT( '/', ColorStr ) + 1 ) )
Pointer := ( ( sFore - 1 ) * LEN( ForeArray() ) ) + sBack + 1
RETURN { Pointer, Pointer }



STATIC FUNCTION ForeArray()
STATIC Array:={'N','B','G','BG','R','RB','GR','W','N+','B+','G+','BG+','R+',;
               'RB+','GR+','W+'}
RETURN Array



STATIC FUNCTION BackArray()
STATIC Array:={'N','B','G','BG','R','RB','GR','W','N*','B*','G*','BG*','R*',;
               'RB*','GR*','W*'}
RETURN Array


  
STATIC FUNCTION ColorText()
STATIC Array:={'Black','Blue','Green','Cyan','Red','Magenta','Brown',;
               'White','Grey','Bright Blue','Bright Green','Bright Cyan',;
               'Bright Red','Bright Magenta','Yellow','Bright White'}
RETURN Array



STATIC FUNCTION CleanUpClr( ColorStr )
LOCAL sFore, sBack, IntensBack
// format the foreground to "string" + intensity indicator "+".
IntensBack := ''
sFore := ForeColor( ColorStr )
sFore := IF( AT('*', sFore) > 0, STRTRAN( sFore, IntensBack := '*', '' ),;
         sFore )
sFore := IF( AT('+', sFore) > 0, STRTRAN( sFore, '+', '' ) + '+', sFore )
// format the background to "string" + intensity indicator "*".
sBack := BackColor( ColorStr )
sBack := STRTRAN( sBack, '+', '*' )
sBack := IF( AT('*', sBack) > 0 .OR. IntensBack == '*',;
         STRTRAN( sBack, '*', '' ) + '*', sBack )
RETURN UPPER( sFore + '/' +  sBack )



STATIC FUNCTION ForeColor( ColorStr )
RETURN SUBSTR( ColorStr, 1, IF( AT( '/', ColorStr ) > 0,;
       AT( '/', ColorStr ), AT( ',', COlorStr ) ) - 1 )



STATIC FUNCTION BackColor( ColorStr )
RETURN IF( AT( '/', ColorStr ) == 0, 'N',;
       SUBSTR( ColorStr, AT( '/', ColorStr ) + 1 ) )



STATIC FUNCTION CheckClrStr( ColorStr )
RETURN ASCAN( ForeArray(), ForeColor( ColorStr ) ) > 0 .AND. ;
       ASCAN( BackArray(), BackColor( ColorStr ) ) > 0



STATIC FUNCTION IDElement( nPaletteID, nItemID )
RETURN ASCAN( Palette( nPaletteID )[3], {|SubArray| nItemID == SubArray[3] } )



STATIC FUNCTION FReadLn( nHandle, nLines, nLineLength )
LOCAL lEndOfFile, cNextBuff, nCount
LOCAL nEOLPos, nEOFPos, nChrsToRead, nFileSize
LOCAL nCurPos, nChrsRead, nPrevPos, cDelimiter
LOCAL lEOF    := .F., cChar   := " ", cBuffer := ""
IF nLines == NIL
  nLines := 1
ENDIF
IF nLineLength == NIL
  nLineLength := 80
ENDIF
cDelimiter := CHR(13) + CHR(10)
nCurPos    := FSEEK( nHandle, 0, FS_RELATIVE )
nFileSize  := FSEEK( nHandle, 0, FS_END )
FSEEK( nHandle, nCurPos, FS_SET )
FOR nCount = 1 TO nLines
  nChrsToRead := nLineLength
  cNextBuff := SPACE( nChrsToRead )
  nPrevPos := FSEEK( nHandle, 0, FS_RELATIVE )
  nChrsRead = FREAD( nHandle, @cNextBuff, nChrsToRead )
  IF ( FSEEK( nHandle, 0, FS_RELATIVE ) == nFileSize )
    IF .NOT. ( AT( cDelimiter, cNextBuff ) < nChrsRead )
       lEOF := .T.
    ENDIF
  ENDIF
  nEOLPos = AT( cDelimiter, cNextBuff ) + 1 // Read past eol?
  IF .NOT. lEOF
    IF nEOLPos > nLineLength                // If cr is nLineLengths char
      FREAD( nHandle, @cChar, 1 )
      cNextBuff = cNextBuff + cChar         // Keep crlf together
      nChrsRead = nChrsRead + 1             // Buffer is nLineLengths + 1 now
    ENDIF
    IF nEOLPos > 1                                 // If eol in buffer
      FSEEK( nHandle, ( nEOLPos - nChrsRead ),;    // Position file pointer
             FS_RELATIVE )
      cNextBuff = SUBSTR( cNextBuff, 1, nEOLPos )  // Keep buffer 1 line
    ENDIF
    IF nEOLPos == 1 
      nCount = nCount - 1
    ENDIF
    cBuffer += cNextBuff
  ENDIF
  IF (lEof)
    EXIT
  ENDIF
NEXT
IF lEOF
  cBuffer := LEFT(cNextBuff, nChrsRead)
ENDIF
RETURN cBuffer



STATIC FUNCTION FEval( nInHandle, bAction, nOutHandle )
STATIC nCurLine := 1
LOCAL cMyBuffer, cResultBuffer
DO WHILE ! cMyBuffer == ""
  cMyBuffer := FReadLn(nInHandle, 1, 256)
  cResultBuffer := EVAL(bAction, cMyBuffer, nCurLine++, nOutHandle)
ENDDO
RETURN cResultBuffer
